#pragma once

#include<dune/pdelab/finiteelement/localbasiscache.hh>
#include<dune/pdelab/localoperator/defaultimp.hh>
#include<dune/pdelab/localoperator/variablefactories.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>

template<typename GFS, typename Container, typename Entity, typename RangeFieldType>
class GFS_LocalViewer
{
public:
  typedef Dune::PDELab::LocalFunctionSpace<GFS> LFS;
  mutable LFS lfs;
  typedef Dune::PDELab::LFSIndexCache<LFS> LFSCache;
  mutable LFSCache lfsCache;
  typedef typename Container::template ConstLocalView<LFSCache> LView;
  mutable LView lView;

  mutable std::vector<RangeFieldType> lv;

  Entity entity;

  GFS_LocalViewer(GFS& gfs, Container& v) :
    lfs(gfs), lfsCache(lfs), lView(v), lv(gfs.maxLocalSize())
    {
    }

  template<typename EG, typename Femspace>
  bool DoSubEvaluation(const EG& eg, const Femspace& femspace)
  {
    if(entity == eg.entity())
    {
      return false;
    }
    entity = eg.entity();

    lfs.bind(entity);
    lfsCache.update();

    lView.bind(lfsCache);
    lView.read_sub_container(femspace,lv);
    lView.unbind();

    return true;
  }

  template<typename EG>
  bool DoEvaluation(const EG& eg)
  {
    if(entity == eg.entity())
    {
      return false;
    }
    entity = eg.entity();

    lfs.bind(entity);
    lfsCache.update();

    lView.bind(lfsCache);
    lView.read(lv);
    lView.unbind();

    return true;
  }

};

//_________________________________________________________________________________________________________________________

template<typename LFS>
class BasisFN
{
public:
  typedef typename LFS::Traits::FiniteElementType::Traits::LocalBasisType::Traits BTraits;
  const std::vector<typename BTraits::RangeType>& phi;
  const std::vector<typename BTraits::JacobianType>& gradphihat;
  std::vector<typename BTraits::JacobianType> gradphi;

  template<typename Pos, typename Cache, typename S>
  BasisFN(const Cache& cache, const Pos& position, const LFS& femspace, S& s) :
    phi(cache.evaluateFunction(position, femspace.finiteElement().localBasis())),
    gradphihat(cache.evaluateJacobian(position, femspace.finiteElement().localBasis())),
    gradphi(femspace.size())
    {
      for (size_t i=0; i<femspace.size(); i++)
      {
        s.mv(gradphihat[i][0],gradphi[i][0]);
      }
    }
};

//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename CHLFS>
class CHVariables
{
public:

  const typename CHLFS::template Child<0>::Type& pfspace;
  const typename CHLFS::template Child<1>::Type& xispace;
  BasisFN<typename CHLFS::template Child<0>::Type> basis;

  RF pf;
  RF xi;
  Dune::FieldVector<RF,dim> gradpf;
  Dune::FieldVector<RF,dim> gradxi;

  template<typename Pos, typename Cache, typename LV, typename S>
  CHVariables(const Cache& cache, const Pos& position, const CHLFS& chLfs, const LV& lv, S& s) :
    pfspace(child(chLfs,Dune::Indices::_0)),
    xispace(child(chLfs,Dune::Indices::_1)),
    basis(cache, position, pfspace, s),
    pf(0.0),
    xi(0.0),
    gradpf(0.0),
    gradxi(0.0)
  {

    for (size_t i=0; i<pfspace.size(); i++)
    {
      pf += lv[pfspace.localIndex(i)]*basis.phi[i];
      xi += lv[xispace.localIndex(i)]*basis.phi[i];
      gradpf.axpy(lv[pfspace.localIndex(i)],basis.gradphi[i][0]);
      gradxi.axpy(lv[xispace.localIndex(i)],basis.gradphi[i][0]);
    }
  }
};


//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename CHLFS>
class CHVariables_PFOld_Extension : public CHVariables<RF,dim,CHLFS>
{
public:


  RF pfOld;
  Dune::FieldVector<RF,dim> gradpfOld;


  template<typename Pos, typename Cache, typename LV, typename S>
  CHVariables_PFOld_Extension(const Cache& cache,
                            const Pos& position, const CHLFS& chLfs, const LV& lv, const LV& lvOld, S& s) :
    CHVariables<RF,dim,CHLFS>(cache, position, chLfs, lv, s),
    pfOld(0.0),
    gradpfOld(0.0)
  {
    for (size_t i=0; i<this->pfspace.size(); i++)
    {
      pfOld += lvOld[this->pfspace.localIndex(i)]*this->basis.phi[i];
      gradpfOld.axpy(lvOld[this->pfspace.localIndex(i)],this->basis.gradphi[i][0]);
    }
  }
};

//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename CHLFS>
class CHVariables_U_PFOld_Extension : public CHVariables<RF,dim,CHLFS>
{
public:
  const typename CHLFS::template Child<2>::Type& uspace;

  RF pfOld;
  Dune::FieldVector<RF,dim> gradpfOld;
  RF u;
  Dune::FieldVector<RF,dim> gradu;
  RF uOld;


  template<typename Pos, typename Cache, typename LV, typename S>
  CHVariables_U_PFOld_Extension(const Cache& cache,
                            const Pos& position, const CHLFS& chLfs, const LV& lv, const LV& lvOld, S& s) :
    CHVariables<RF,dim,CHLFS>(cache, position, chLfs, lv, s),
    uspace(child(chLfs,Dune::Indices::_2)),
    pfOld(0.0),
    gradpfOld(0.0),
    u(0.0),
    gradu(0.0),
    uOld(0.0)
  {
    for (size_t i=0; i<this->pfspace.size(); i++)
    {
      u += lv[uspace.localIndex(i)]*this->basis.phi[i];
      gradu.axpy(lv[uspace.localIndex(i)],this->basis.gradphi[i][0]);
      uOld += lvOld [uspace.localIndex(i)]*this->basis.phi[i];
      pfOld += lvOld[this->pfspace.localIndex(i)]*this->basis.phi[i];
      gradpfOld.axpy(lvOld[this->pfspace.localIndex(i)],this->basis.gradphi[i][0]);
    }
  }
};


//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename CHLFS>
class CHVariables_Threephase_Extension : public CHVariables<RF,dim,CHLFS>
{
public:
  const typename CHLFS::template Child<2>::Type& pf2space;
  const typename CHLFS::template Child<3>::Type& xi2space;

  RF pf2;
  Dune::FieldVector<RF,dim> gradpf2;
  RF xi2;
  Dune::FieldVector<RF,dim> gradxi2;

  template<typename Pos, typename Cache, typename LV, typename S>
  CHVariables_Threephase_Extension(const Cache& cache,
                            const Pos& position, const CHLFS& chLfs, const LV& lv, S& s) :
    CHVariables<RF,dim,CHLFS>(cache, position, chLfs, lv, s),
    pf2space(child(chLfs,Dune::Indices::_2)),
    xi2space(child(chLfs,Dune::Indices::_3)),
    pf2(0.0),
    gradpf2(0.0),
    xi2(0.0),
    gradxi2(0.0)
  {
    for (size_t i=0; i<this->pfspace.size(); i++)
    {
      pf2 += lv[pf2space.localIndex(i)]*this->basis.phi[i];
      gradpf2.axpy(lv[pf2space.localIndex(i)],this->basis.gradphi[i][0]);
      xi2 += lv[xi2space.localIndex(i)]*this->basis.phi[i];
      gradxi2.axpy(lv[xi2space.localIndex(i)],this->basis.gradphi[i][0]);
    }
  }
};


//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename CHLFS>
class CHVariables_Threephase_PFOld_Extension : public CHVariables_Threephase_Extension<RF,dim,CHLFS>
{
public:

  RF pfOld;
  Dune::FieldVector<RF,dim> gradpfOld;
  RF pf2Old;
  Dune::FieldVector<RF,dim> gradpf2Old;


  template<typename Pos, typename Cache, typename LV, typename S>
  CHVariables_Threephase_PFOld_Extension(const Cache& cache,
                            const Pos& position, const CHLFS& chLfs, const LV& lv, const LV& lvOld, S& s) :
    CHVariables_Threephase_Extension<RF,dim,CHLFS>(cache, position, chLfs, lv, s),
    pfOld(0.0),
    gradpfOld(0.0),
    pf2Old(0.0),
    gradpf2Old(0.0)
  {
    for (size_t i=0; i<this->pfspace.size(); i++)
    {
      pfOld += lvOld[this->pfspace.localIndex(i)]*this->basis.phi[i];
      gradpfOld.axpy(lvOld[this->pfspace.localIndex(i)],this->basis.gradphi[i][0]);
      pf2Old += lvOld[this->pf2space.localIndex(i)]*this->basis.phi[i];
      gradpf2Old.axpy(lvOld[this->pf2space.localIndex(i)],this->basis.gradphi[i][0]);
    }
  }
};


template<typename RF, const int dim, typename CHLFS>
class CHVariables_Threephase_U_PFOld_Extension : public CHVariables_Threephase_Extension<RF,dim,CHLFS>
{
public:
  const typename CHLFS::template Child<4>::Type& uspace;

  RF pfOld;
  Dune::FieldVector<RF,dim> gradpfOld;
  RF pf2Old;
  Dune::FieldVector<RF,dim> gradpf2Old;
  RF u;
  Dune::FieldVector<RF,dim> gradu;
  RF uOld;


  template<typename Pos, typename Cache, typename LV, typename S>
  CHVariables_Threephase_U_PFOld_Extension(const Cache& cache,
                            const Pos& position, const CHLFS& chLfs, const LV& lv, const LV& lvOld, S& s) :
    CHVariables_Threephase_Extension<RF,dim,CHLFS>(cache, position, chLfs, lv, s),
    uspace(child(chLfs,Dune::Indices::_4)),
    pfOld(0.0),
    gradpfOld(0.0),
    pf2Old(0.0),
    gradpf2Old(0.0),
    u(0.0),
    gradu(0.0),
    uOld(0.0)
  {
    for (size_t i=0; i<this->pfspace.size(); i++)
    {
      u += lv[uspace.localIndex(i)]*this->basis.phi[i];
      gradu.axpy(lv[uspace.localIndex(i)],this->basis.gradphi[i][0]);
      uOld += lvOld [uspace.localIndex(i)]*this->basis.phi[i];
      pfOld += lvOld[this->pfspace.localIndex(i)]*this->basis.phi[i];
      gradpfOld.axpy(lvOld[this->pfspace.localIndex(i)],this->basis.gradphi[i][0]);
      pf2Old += lvOld[this->pf2space.localIndex(i)]*this->basis.phi[i];
      gradpf2Old.axpy(lvOld[this->pf2space.localIndex(i)],this->basis.gradphi[i][0]);
    }
  }
};


//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename NSLFS>
class NSVariables
{
public:
  const typename NSLFS::template Child<0>::Type& vspace;
  const typename NSLFS::template Child<0>::Type::template Child<0>::Type& vfemspace;
  BasisFN<typename NSLFS::template Child<0>::Type::template Child<0>::Type> vbasis;

  Dune::FieldVector<RF,dim> v;
  Dune::FieldMatrix<RF,dim,dim> jacv;
  RF div_v;

  template<typename Pos, typename vCache, typename LV, typename S>
  NSVariables(const vCache& vcache, const Pos& position, const NSLFS& nsLfs, const LV& lv, S& s) :
    vspace(child(nsLfs,Dune::Indices::_0)),
    vfemspace(child(vspace,Dune::Indices::_0)),
    vbasis(vcache, position, vfemspace, s),
    v(0.0),
    jacv(0.0),
    div_v(0.0)
  {
    for(size_t k=0; k<dim; k++)
    {
      for (size_t i=0; i<vfemspace.size(); i++)
      {
        v[k] += lv[child(vspace,k).localIndex(i)]*vbasis.phi[i];
        jacv[k].axpy(lv[child(vspace,k).localIndex(i)],vbasis.gradphi[i][0]);
      }
      div_v += jacv[k][k];  // div is trace of jacobian
    }
  }
};
//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename NSLFS>
class NSVariables_P_Extension : public NSVariables<RF,dim,NSLFS>
{
public:
  const typename NSLFS::template Child<1>::Type& pspace;
  BasisFN<typename NSLFS::template Child<1>::Type> pbasis;

  RF p;
  Dune::FieldVector<RF,dim> gradp;

  template<typename Pos, typename vCache, typename pCache, typename LV, typename S>
  NSVariables_P_Extension(const vCache& vcache, const pCache& pcache,
                            const Pos& position, const NSLFS& nsLfs, const LV& lv, S& s) :
    NSVariables<RF,dim,NSLFS>(vcache, position, nsLfs, lv, s),
    pspace(child(nsLfs,Dune::Indices::_1)),
    pbasis(pcache, position, pspace, s),
    p(0.0),
    gradp(0.0)
  {
    for (size_t i=0; i<pspace.size(); i++)
    {
      p += lv[pspace.localIndex(i)]*pbasis.phi[i];
      gradp.axpy(lv[pspace.localIndex(i)],pbasis.gradphi[i][0]);
    }
  }
};

//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename NSLFS>
class NSVariables_P_VOld_Extension : public NSVariables_P_Extension<RF,dim,NSLFS>
{
public:
  Dune::FieldVector<RF,dim> vOld;

  template<typename Pos, typename vCache, typename pCache, typename LV, typename S>
  NSVariables_P_VOld_Extension(const vCache& vcache, const pCache& pcache,
                            const Pos& position, const NSLFS& nsLfs, const LV& lv, const LV& lvOld, S& s) :
    NSVariables_P_Extension<RF,dim,NSLFS>(vcache, pcache, position, nsLfs, lv, s),
    vOld(0.0)
  {

    for (size_t i=0; i<this->vfemspace.size(); i++)
    {
      for(size_t k=0; k<dim; k++)
      {
        vOld[k] += lvOld[child(this->vspace,k).localIndex(i)]*this->vbasis.phi[i];
      }
    }
  }
};
//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename CHNSLFS>
class CHNSVariables_All
{
public:
  const typename CHNSLFS::template Child<1>::Type& nsspace;
  const typename CHNSLFS::template Child<1>::Type::template Child<0>::Type& vspace;
  const typename CHNSLFS::template Child<1>::Type::template Child<0>::Type::template Child<0>::Type& vfemspace;
  BasisFN<typename CHNSLFS::template Child<1>::Type::template Child<0>::Type::template Child<0>::Type> vbasis;
  Dune::FieldVector<RF,dim> v;
  Dune::FieldMatrix<RF,dim,dim> jacv;
  RF div_v;

  const typename CHNSLFS::template Child<1>::Type::template Child<1>::Type& pspace;
  BasisFN<typename CHNSLFS::template Child<1>::Type::template Child<1>::Type> pbasis;
  RF p;
  Dune::FieldVector<RF,dim> gradp;

  const typename CHNSLFS::template Child<0>::Type& chspace;
  const typename CHNSLFS::template Child<0>::Type::template Child<0>::Type& pfspace;
  const typename CHNSLFS::template Child<0>::Type::template Child<1>::Type& xispace;
  const typename CHNSLFS::template Child<0>::Type::template Child<2>::Type& pf2space;
  const typename CHNSLFS::template Child<0>::Type::template Child<3>::Type& xi2space;
  BasisFN<typename CHNSLFS::template Child<0>::Type::template Child<0>::Type> basis;
  RF pf;
  RF pf2;
  RF xi;
  RF xi2;
  Dune::FieldVector<RF,dim> gradpf;
  Dune::FieldVector<RF,dim> gradxi;
  Dune::FieldVector<RF,dim> gradpf2;
  Dune::FieldVector<RF,dim> gradxi2;

  const typename CHNSLFS::template Child<0>::Type::template Child<4>::Type& uspace;
  RF u;
  Dune::FieldVector<RF,dim> gradu;


  template<typename Pos, typename CHCache, typename VCache, typename PCache, typename LV, typename S>
  CHNSVariables_All(const CHCache& chcache, const VCache& vcache, const PCache& pcache,
                    const Pos& position, const CHNSLFS& chnsLfs, const LV& lv, S& s) :
    nsspace(child(chnsLfs,Dune::Indices::_1)),
    vspace(child(nsspace,Dune::Indices::_0)),
    vfemspace(child(vspace,Dune::Indices::_0)),
    vbasis(vcache, position, vfemspace, s),
    v(0.0),
    jacv(0.0),
    div_v(0.0),
    pspace(child(nsspace,Dune::Indices::_1)),
    pbasis(pcache, position, pspace, s),
    p(0.0),
    gradp(0.0),
    chspace(child(chnsLfs,Dune::Indices::_0)),
    pfspace(child(chspace,Dune::Indices::_0)),
    xispace(child(chspace,Dune::Indices::_1)),
    pf2space(child(chspace,Dune::Indices::_2)),
    xi2space(child(chspace,Dune::Indices::_3)),
    basis(chcache, position, pfspace, s),
    pf(0.0),
    pf2(0.0),
    xi(0.0),
    xi2(0.0),
    gradpf(0.0),
    gradxi(0.0),
    gradpf2(0.0),
    gradxi2(0.0),
    uspace(child(chspace,Dune::Indices::_4)),
    u(0.0),
    gradu(0.0)
  {
    for (size_t i=0; i<pfspace.size(); i++)
    {
      pf += lv[pfspace.localIndex(i)]*basis.phi[i];
      pf2 += lv[pf2space.localIndex(i)]*basis.phi[i];
      xi += lv[xispace.localIndex(i)]*basis.phi[i];
      xi2 += lv[xi2space.localIndex(i)]*basis.phi[i];
      gradpf.axpy(lv[pfspace.localIndex(i)],basis.gradphi[i][0]);
      gradxi.axpy(lv[xispace.localIndex(i)],basis.gradphi[i][0]);
      gradpf2.axpy(lv[pf2space.localIndex(i)],basis.gradphi[i][0]);
      gradxi2.axpy(lv[xi2space.localIndex(i)],basis.gradphi[i][0]);
      u += lv[uspace.localIndex(i)]*basis.phi[i];
      gradu.axpy(lv[uspace.localIndex(i)],basis.gradphi[i][0]);
    }
    for(size_t k=0; k<dim; k++)
    {
      for (size_t i=0; i<vfemspace.size(); i++)
      {
        v[k] += lv[child(vspace,k).localIndex(i)]*vbasis.phi[i];
        jacv[k].axpy(lv[child(vspace,k).localIndex(i)],vbasis.gradphi[i][0]);
      }
      div_v += jacv[k][k];  // div is trace of jacobian
    }
    for (size_t i=0; i<pspace.size(); i++)
    {
      p += lv[pspace.localIndex(i)]*pbasis.phi[i];
      gradp.axpy(lv[pspace.localIndex(i)],pbasis.gradphi[i][0]);
    }
  }
};

//_________________________________________________________________________________________________________________________

// Local Basis Caches used for Operator Splitting

template<typename CHGFS>
using CHLocalBasisCache =
  Dune::PDELab::LocalBasisCache<typename CHGFS::template Child<0>::Type::Traits::FiniteElementType::Traits::LocalBasisType>;


template<typename NSGFS>
using VLocalBasisCache =
  Dune::PDELab::LocalBasisCache<typename NSGFS::template Child<0>::Type::template Child<0>::Type::Traits::FiniteElementType::Traits::LocalBasisType>;

template<typename NSGFS>
using PLocalBasisCache =
  Dune::PDELab::LocalBasisCache<typename NSGFS::template Child<1>::Type::Traits::FiniteElementType::Traits::LocalBasisType>;


// Local Basis Caches used for Monolithic Structure

template<typename CHNSGFS>
using CH_CHNSLocalBasisCache =
  Dune::PDELab::LocalBasisCache<typename CHNSGFS::template Child<0>::Type::template Child<0>::Type::Traits::FiniteElementType::Traits::LocalBasisType>;

template<typename CHNSGFS>
using V_CHNSLocalBasisCache =
  Dune::PDELab::LocalBasisCache<typename CHNSGFS::template Child<1>::Type::template Child<0>::Type::template Child<0>::Type::Traits::FiniteElementType::Traits::LocalBasisType>;

template<typename CHNSGFS>
using P_CHNSLocalBasisCache =
  Dune::PDELab::LocalBasisCache<typename CHNSGFS::template Child<1>::Type::template Child<1>::Type::Traits::FiniteElementType::Traits::LocalBasisType>;
