#pragma once

// C includes
#include<sys/stat.h>
// C++ includes
#include<iostream>
// Dune includes
#include<dune/grid/io/file/vtk.hh>
#include<dune/pdelab/newton/newton.hh>

#include "thinstrip_full_chns_monolithic.hh"
#include <dune/phasefield/localoperator/chns_monolithic_diff_estimator.hh>

#include <dune/phasefield/utility/ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES.hh>

#include "thinstrip_parameters.hh"
#include "thinstrip_boundary.hh"
#include <dune/phasefield/chns_base.hh>
#include <dune/phasefield/ffs_chns_r.hh>
#include <dune/phasefield/initial_fn.hh>
#include <dune/phasefield/debug/vector_debug.hh>



//_________________________________________________________________________________________________________________________

template<typename Grid, typename GV, typename ES, typename V_FEM, typename P_FEM, typename PHI_FEM>
void driver_3p(Grid& grid, GV& gv, ES& es, const V_FEM& vFem, const P_FEM pFem, const PHI_FEM& phiFem, Dune::ParameterTree& ptree, Dune::MPIHelper& helper)
{

  const int dim = GV::dimension;
  typedef double RF;


  typedef Params_thinstrip< RF,
                        //TripleWell<RF,DoubleWell_limited< RF, DoubleWell_poly<RF>, Limiter_FakeDiverging<RF> > >,
                        TripleWell<RF,DoubleWell_limited< RF, DoubleWell_8th_order_poly<RF>, Limiter_FakeDiverging<RF> > >,
                        Mobility_Const<RF>,
                        Reaction_Thinstrip<RF>,
                        VelDissipation_Quadratic_Shifted<RF> > Parameters;
  Parameters param(ptree);
  TimestepManager timestepManager(ptree.sub("Time"),param.time,2);


//_________________________________________________________________________________________________________________________
// Make grid function spaces


  typedef GFS_ffs_chns_r_monolithic<RF, ES, V_FEM, P_FEM, PHI_FEM, PHI_FEM> GFS;
  GFS gfs(es,vFem,pFem,phiFem,phiFem);
  typedef typename GFS::CHNS CHNS_GFS;


  //_________________________________________________________________________________________________________________________
  //Constraints
  typedef DomainSize<RF> DS;
  DS dS(0,1,0,ptree.get("domain.len",1),1e-8);
  typedef Bdry_ffs_chns_r_compositeV_2d_Explicit_Domain_monolithic<
              BottomBoundary<DS>,            // v_x
              AllBoundary<DS>,         // v_y
              NoBoundary<DS>,                                      // p
              NoBoundary<DS>,                                    // phi1
              NoBoundary<DS>,       // mu1
              NoBoundary<DS>,                                    // phi2
              NoBoundary<DS>,       // mu2
              NoBoundary<DS>>                                      // u
              Bdry;
  Bdry bdry(dS);
  gfs.updateConstraintsContainer(bdry);

  //_________________________________________________________________________________________________________________________
  //Coefficient vectors

  using CHNS_V = Dune::PDELab::Backend::Vector<CHNS_GFS, RF>;
  CHNS_V chnsV(gfs.chns);

  //_________________________________________________________________________________________________________________________
  //Initial Values

    typedef Ini_ffs_chns_r_monolithic< VelocityInitial_InOutFlow<GV,RF,Parameters,dim>,
                      PInitial_InOutFlow<GV,RF,Parameters>,
                      PhiInitial_Thinstrip_Layers<GV,RF,Parameters,0>,
                      ZeroInitial<GV,RF,Parameters>,
                      PhiInitial_Thinstrip_Layers<GV,RF,Parameters,1>,
                      ZeroInitial<GV,RF,Parameters>,
                      UInitial_Inflow<GV,RF,Parameters> >  Initial;
    Initial initial(gv,param);
    Dune::PDELab::interpolate(initial.iniCHNS, gfs.chns, chnsV);

  //_________________________________________________________________________________________________________________________
// Refine Intial Values
    for(int i=0; i<ptree.get("RefinementInitial.maxRefineLevel",2); i++)
    {
      // mark elements for refinement


      typedef RefinementIndicator<GV,RF,CHNS_MONOLITHIC_DIFF_ESTIMATOR<CHNS_GFS>,CHNS_GFS> RIND0;
      RIND0 rInd0(gv,gfs.chns,ptree.sub("RefinementInitial"));
      rInd0.calculateIndicator(gv,chnsV,false);
      rInd0.markGrid(grid,2);
      rInd0.myRefineGrid(grid,gv,chnsV,gfs.chns,1);
      Dune::PDELab::interpolate(initial.iniCHNS , gfs.chns, chnsV);
    }

    gfs.updateConstraintsContainer(bdry);
    interpolateToBdry_monolithic(gfs,initial.iniCHNS,chnsV);

    CHNS_V chnsVOld = chnsV;

  //_________________________________________________________________________________________________________________________
  // LocalOperators



  typedef THINSTRIP_FULL_CHNS_Monolithic<Parameters, CHNS_GFS, CHNS_V> CHNS_LOP;
  CHNS_LOP chnsLop(param, gfs.chns, chnsVOld);

  typedef Dune::PDELab::ISTL::BCRSMatrixBackend<> MBE;
  MBE mbe(10);// Maximal number of nonzeroes per row can be cross-checked by patternStatistics().


  //_________________________________________________________________________________________________________________________
  // Construct discrete Grid Functions
  typedef DGF_ffs_chns_r_monolithic<GFS, CHNS_V> DGF;
  DGF dgf(gfs, chnsV);

  //_________________________________________________________________________________________________________________________
  // prepare VTK writer
  std::string filename=ptree.get("output.filename","output");
  struct stat st;
  if( stat( filename.c_str(), &st ) != 0 )
  {
    int stat = 0;
    stat = mkdir(filename.c_str(),S_IRWXU|S_IRWXG|S_IRWXO);
    if( stat != 0 && stat != -1)
      std::cout << "Error: Cannot create directory "  << filename << std::endl;
  }
  auto stationaryvtkwriter = std::make_shared<Dune::SubsamplingVTKWriter<GV> >(gv,0);
  Dune::VTKSequenceWriter<GV> vtkwriter(stationaryvtkwriter,filename.c_str(),filename.c_str(),"");
  dgf.addToVTKWriter(vtkwriter);

  vtkwriter.write(param.time.t,Dune::VTK::ascii);

//_________________________________________________________________________________________________________________________
  printVector(chnsV,20,"chnsV");


  //____________________________________________________________________________________________________________________
  //int i_it=0;
  chnsVOld = chnsV;
  while (!timestepManager.isFinished())
    {

      // Grid Operators
      typedef Dune::PDELab::GridOperator<CHNS_GFS,CHNS_GFS,CHNS_LOP,MBE,RF,RF,RF,typename GFS::CHNS_CC,typename GFS::CHNS_CC> CHNS_GO;
      CHNS_GO chnsGo(gfs.chns,gfs.chnsCC,gfs.chns,gfs.chnsCC,chnsLop,mbe);

      // Linear solver
      typedef ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES
                  <CHNS_GO, Dune::SeqILU0, Dune::RestartedGMResSolver> CHNS_LS; CHNS_LS ls_chns(chnsGo,1200,200,3,1);
      //typedef Dune::PDELab::ISTLBackend_SEQ_GMRES_ILU0 CHNS_LS;  CHNS_LS ls_chns(200,5000,1);

      // Nonlinear solver
      typedef Dune::PDELab::Newton<CHNS_GO,CHNS_LS,CHNS_V> PDESolver1;
      PDESolver1 chns_newton(chnsGo,chnsV,ls_chns);
      chns_newton.setParameters(ptree.sub("NS_Newton"));

      //_________________________________________________________________________________________________________________________
      // do time step


      interpolateToBdry_monolithic(gfs,initial.iniCHNS, chnsV);

      int result = timestepManager.doOnlyNSTimestep(chns_newton,vtkwriter);
      if(result <0)
      {
        return;
      }
      else if(result == 0)
      {
        chnsV = chnsVOld;
        continue;
      }

      chnsVOld = chnsV;
//TODO?
//child(initial.iniCH,Dune::Indices::_0).ApplyReactionDt(param);
      //_________________________________________________________________________________________________________________________

      // mark elements for refinement

      typedef RefinementIndicator<GV,RF,CHNS_MONOLITHIC_DIFF_ESTIMATOR<CHNS_GFS>,CHNS_GFS> RIND;
      RIND rInd(gv,gfs.chns,ptree.sub("Refinement"));
      rInd.calculateIndicator(gv,chnsV,false);
      rInd.markGrid(grid,2);
      rInd.myRefineGrid(grid,gv,chnsV,chnsVOld,gfs.chns,1);

      //vtkwriter.write(param.time.t+1e-6,Dune::VTK::ascii);

      // recompute constraints
      gfs.updateConstraintsContainer(bdry);
      interpolateToBdry_monolithic(gfs,initial.iniCHNS,chnsV);

      //vtkwriter.write(timestepManager.t+2e-6,Dune::VTK::ascii);
    }
}
