#pragma once

#include<dune/pdelab/common/quadraturerules.hh>
#include<dune/pdelab/finiteelement/localbasiscache.hh>
#include<dune/pdelab/localoperator/defaultimp.hh>
#include<dune/pdelab/localoperator/pattern.hh>
#include<dune/pdelab/localoperator/flags.hh>
#include<dune/pdelab/localoperator/variablefactories.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>

#include<dune/phasefield/localoperator/fem_base_chns.hh>

//Parameter class (for functions), FEM space and RangeField
template<typename Param, typename CHNSGFS, typename CHNSContainer>
class THINSTRIP_FULL_CHNS_Monolithic :
  public Dune::PDELab::FullVolumePattern,
  public Dune::PDELab::LocalOperatorDefaultFlags,
  public Dune::PDELab::NumericalJacobianVolume<THINSTRIP_FULL_CHNS_Monolithic<Param, CHNSGFS, CHNSContainer> >,
  public Dune::PDELab::NumericalJacobianApplyVolume<THINSTRIP_FULL_CHNS_Monolithic<Param, CHNSGFS, CHNSContainer> >
{
private:
  typedef typename CHNSContainer::field_type RF;
  typedef typename CHNSGFS::Traits::GridViewType::template Codim<0>::Entity   ENTITY;

  CH_CHNSLocalBasisCache<CHNSGFS> chCache;
  V_CHNSLocalBasisCache<CHNSGFS> vCache;
  P_CHNSLocalBasisCache<CHNSGFS> pCache;

  typedef GFS_LocalViewer<CHNSGFS, CHNSContainer, ENTITY, RF> CHNSLView;
  mutable CHNSLView chnsOldLocal;

  Param& param; // parameter functions

public:
  enum {doPatternVolume=true};
  //enum {doLambdaVolume=true};
  enum {doAlphaVolume=true};

  //constructor
  THINSTRIP_FULL_CHNS_Monolithic(Param& param_, CHNSGFS& chnsGfs_, CHNSContainer& chnsVOld_) :
    chnsOldLocal(chnsGfs_,chnsVOld_),
    param(param_)
  {
  }


  template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
  void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
  {
    //Quadrature Rule
    const int dim = EG::Entity::dimension;
    const int order = 2; //TODO: Choose more educated
    auto rule = Dune::PDELab::quadratureRule(eg.geometry(), order);

    chnsOldLocal.DoEvaluation(eg);

    for(const auto& ip : rule)
    {
      const auto S = eg.geometry().jacobianInverseTransposed(ip.position());
      auto factor = ip.weight()*eg.geometry().integrationElement(ip.position());

      CHNSVariables_All<RF,dim,LFSV> chns(chCache, vCache, pCache, ip.position(), lfsv, x.base(), S);
      CHNSVariables_All<RF,dim,LFSV> chnsOld(chCache, vCache, pCache, ip.position(), lfsv, chnsOldLocal.lv, S);

      RF pf3 = 1-chns.pf-chns.pf2;
      RF xi3 = -param.tw.Sigma3*(chns.xi/param.tw.Sigma1 + chns.xi2/param.tw.Sigma2);
      RF pf_f= chns.pf+chns.pf2+2*param.delta*pf3;

      RF pf_c = chns.pf;
      RF pf_c_Old = chnsOld.pf;
      RF pf_c_reg = pf_c+param.delta;
      RF pf_c_reg_Old = pf_c_Old + param.delta;


      Dune::FieldVector<RF,dim> gradpf_f(0.0);
      Dune::FieldVector<RF,dim> outerGradP(0.0);
      outerGradP[0]=-param.gradP;
      gradpf_f.axpy((1-2*param.delta),chns.gradpf+chns.gradpf2);

      RF rho_f=param.phys.rho_f(chns.pf,chns.pf2);
      RF rho_f_reg = rho_f + (param.phys.rho_1+param.phys.rho_2)*param.delta;

      RF source = 0;
      Dune::FieldVector<RF, dim> xg = eg.geometry().global( ip.position() );
      if(xg[0] > 0.1 && xg[0] < 0.3)
      {
        source = (xg[0]-0.1)*(0.3-xg[0])/0.04;
      }

      for (size_t i=0; i<chns.pfspace.size(); i++)
      {
        //RF Rspeed = std::max(param.tw.eval_q(ch.pf,ch.pf2,pf3) - 0.1, 0.0);
        RF Rspeed = param.tw.eval_q(chns.pf,chns.pf2,pf3);
        RF R1 = -Rspeed/param.eps*
                  (param.reaction.eval(chns.u)+param.phys.curvatureAlpha*(chns.xi-xi3))*chns.basis.phi[i];
        RF J1 = param.mobility.eval(chns.pf,chns.pf2,pf3)*param.eps/param.tw.Sigma1 * (chns.gradxi*chns.basis.gradphi[i][0]);

      // integrate: (pf-pfold)*phi_i/delta_t + gradpf * grad phi_i + ...
        r.accumulate(chns.pfspace,i,factor*(
          (chns.pf-chnsOld.pf)*chns.basis.phi[i]/param.time.dt
          +J1
          +(chns.pf*chns.div_v + (chns.gradpf*chns.v))*chns.basis.phi[i] //TODO div_v or div_v OLD?, v or vOld
          //-pf1*(v*gradphi[i][0])   weak transport
          -R1
        ));

        r.accumulate(chns.pf2space,i,factor*(
          (chns.pf2-chnsOld.pf2)*chns.basis.phi[i]/param.time.dt
          +param.mobility.eval(chns.pf,chns.pf2,pf3)*param.eps/param.tw.Sigma2 * (chns.gradxi2*chns.basis.gradphi[i][0])
          +(chns.pf2*chns.div_v + (chns.gradpf2*chns.v))*chns.basis.phi[i]      //TODO div_v or div_v OLD?, v or vOld
          //-pf2*(v*gradphi[i][0])
        ));

        r.accumulate(chns.xispace,i,factor*(
          chns.xi*chns.basis.phi[i]
          - 1/param.eps* param.tw.eval_dpf1(chns.pf,chns.pf2,pf3) *chns.basis.phi[i]
          - param.eps * param.tw.Sigma1* (chns.gradpf*chns.basis.gradphi[i][0])
        ));

        r.accumulate(chns.xi2space,i,factor*(
          chns.xi2*chns.basis.phi[i]
          - 1/param.eps* param.tw.eval_dpf2(chns.pf,chns.pf2,pf3) *chns.basis.phi[i]
          - param.eps * param.tw.Sigma2* (chns.gradpf2*chns.basis.gradphi[i][0])
        ));

      // integrate (pf-pfold)*(u-uast)*phi_i/delta_t + pfold*(u-uold)*phi_i/delta_t + D*pf*gradu*grad phi_i = 0

//      r.accumulate(chns.uspace,i,factor*(chns.u-param.phys.uAst)*chns.basis.phi[i]);



        //Dune::FieldVector<RF, dim> xg = eg.geometry().global( ip.position() );
        r.accumulate(chns.uspace,i,factor*(

          //(u-param.uAst)*phi[i]/param.dt //For Testing

          (pf_c_reg-pf_c_reg_Old)*(chns.u-param.phys.uAst)*chns.basis.phi[i]/param.time.dt
          + pf_c_reg_Old*(chns.u-chnsOld.u)*chns.basis.phi[i]/param.time.dt
          + param.phys.D*pf_c_reg_Old*(chns.gradu*chns.basis.gradphi[i][0])
          + J1*(chns.u-param.phys.uAst)
          //- pf_c*(u-param.phys.uAst)*(v*gradphi[i][0])
          + (pf_c*(chns.u-param.phys.uAst)*chns.div_v                            //TODO v or vOld
              + (chns.u-param.phys.uAst)*(chns.gradpf*chns.v)
              + pf_c*(chns.gradu*chns.v)                        )*chns.basis.phi[i]
          - pf_c
          ));
      } // for i

/*      for (size_t i=0; i<chns.pspace.size(); i++)
      {
        for(int k=0; k<dim; k++)
        {
          r.accumulate(chns.pspace,i, -1.0 * (chns.p )* chns.pbasis.phi[i] * factor);
        }
      }
      for (size_t i=0; i<chns.vfemspace.size(); i++)
      {
        for(int k=0; k<dim; k++)
        {
          r.accumulate(child(chns.vspace,k),i, (   chns.v[k]*chns.vbasis.phi[i]      )*factor);
        } //for k
      } // for i
*/


      for (size_t i=0; i<chns.pspace.size(); i++)
      {
        for(int k=0; k<dim; k++)
        {
          // integrate: (div v)*phi
          //r.accumulate(ns.pspace,i, -1.0* pf_f*(v*gradphi[i][0])*factor);            Weak form
          r.accumulate(chns.pspace,i, -1.0 * (pf_f*chns.div_v  + (gradpf_f*chns.v) )* chns.pbasis.phi[i] * factor);
        }
      }
      for (size_t i=0; i<chns.vfemspace.size(); i++)
      {
        for(int k=0; k<dim; k++)
        {
          // integrate: ( (rho+rhoOld)/2 * v_n+1 - rhoOld * v_n)/dt * phi   + nu * Dv_n+1 : D phi - p_n+1 div phi

          //std::cout << (param.nu * (jacv[k]*gradvphi[i][0]) - p * gradvphi[i][0][k]
          //- param.rho*g[k]*vphi[i] )*factor << std::endl;

          //RF lambda = 3;
          //const auto v_nabla_v = v * jacv[k];
          RF gradxi3k = -param.tw.Sigma3*(chns.gradxi[k]/param.tw.Sigma1 + chns.gradxi2[k]/param.tw.Sigma2);
          RF S = 0;
          S += - chns.xi2*(chns.gradpf[k]*pf_f  - chns.pf  * gradpf_f[k])/(pf_f*pf_f); //TODO!!!!!!! Other pf_f scaling!!!!!
          S += - chns.xi *(chns.gradpf2[k]*pf_f - chns.pf2 * gradpf_f[k])/(pf_f*pf_f);
          S += - 2*param.delta*pf3*(gradxi3k-chns.gradxi2[k]-chns.gradxi[k]);

          //Stronger form - flow in interfaces possible
          //S += -ch.pf*ch.gradxi[k] - ch.pf2*ch.gradxi2[k] - 2*param.delta*pf3*gradxi3k;

          //Works but is not the one in the theory (same as above, w/o regularization)
          //S += -ch.pf*ch.gradxi[k] - ch.pf2*ch.gradxi2[k];

          S = S*param.SurfaceTensionScaling*chns.vbasis.phi[i];

          r.accumulate(child(chns.vspace,k),i, (   rho_f_reg*(chns.v[k]-chnsOld.v[k])*chns.vbasis.phi[i]/param.time.dt          //TimeEvolution
                                            //- ns.p * (pf_f*ns.vbasis.gradphi[i][0][k] + ns.vbasis.phi[i]*gradpf_f[k])            //Pressure in weak form
                                            + pf_f * ((chns.gradp[k]+outerGradP[k]) * chns.vbasis.phi[i]) //Pressure in strong form
                                            + param.phys.eval_mu(pf_f) * (chns.jacv[k]*chns.vbasis.gradphi[i][0]) //+vphi[i]*(jacv[k]*gradpf)   )  //Viscosity
                                            + param.vDis.eval(pf_f)*chns.v[k]*chns.vbasis.phi[i]              //Velocity Dissipation in Solid
                                            //- rho*g[k]*vphi[i]                //Gravitation
                                            - S                                //Surface Tension
                                          )*factor);
          if(k==0)
          {
            r.accumulate(child(chns.vspace,k),i, ( pf_f * (param.phys.outerMomentumXForce * chns.vbasis.phi[i])   // Momentum Forcing Term
                                          )*factor);
          }
        } //for k
      } // for i

    } // for ip
  }

};
